#############################################################################
##                                                                         ##
##  basics.gi                       PTREE                  Alexander Cant  ##
##                                                                         ##
#############################################################################

#############################################################################
##
##  Input:
##    p     a prime
##
##  Output:
##    The integral version of the p-adic space group as PcpGroup.
##
##  The integral version of the p-adic space group has generators g1,...,gp
##  and it is a split extension of a cyclic group P = <g1> of size p by the
##  normal subgroup T = <g2,...,gp> that is free abelian of rank p-1.
##  Here we think of
##  (a) the generator of P as a primitive p-th root of unity over QQ, and of
##  (b) T as the free ZZ-module ZZ[g1] of rank p-1.
##  The generators g2,...,gp of T represent the elements of the basis
##  {1,g1,g1^2,...,g1^{p-2}}, and P acts on T via multiplication, i.e.
##      gi^g1 = g{i+1} for 1 < i < p  and  gp^g1 = g2^-1 ... gp^-1.
##
SpaceGroupMaxClass := function( p )
    local coll, i, k;

    # initialize collector
    coll := FromTheLeftCollector( p );

    # set relative order of first generator to p (default value is 0)
    SetRelativeOrder( coll, 1, p );

    # set gi^g1 to g{i+1} for 1 < i < p
    for i in [2..p-1] do
        SetConjugate( coll, i, 1, [i+1,1] );
    od;

    # set gp^g1 to g2^-1 ... gp^-1
    k := ListWithIdenticalEntries( 2*p-2, -1 );
    for i in [1..p-1] do
        k[2*i-1] := i+1;
    od;
    SetConjugate( coll, p, 1, k );

    # compute remaining relations
    UpdatePolycyclicCollector( coll );

    # return the pcp group that is defined by coll
    return PcpGroupByCollector( coll );
end;


#############################################################################
##
##  Input:
##    p     a prime
##    n     an integer > 1
##
##  Output:
##    The unique main-line group of size p^n as a PcGroup.
##
MainLineGroup := function( p, n )
    local S, F, G;

    # check p and n
    if not ( IsInt( p ) and IsPrime( p ) and IsInt( n ) and n > 1 ) then
        Error( "1st argument has to be a prime and ",
                "2nd argument an integer > 1" );
        return fail;
    fi;

    # get space group
    S := SpaceGroupMaxClass( p );
    F := PcpGroupToFpGroup( S );

    # get main-line group of size p^n as p-quotient of F
    G := Pq( F : Prime := p, ClassBound := n-1 );

    return G;
end;


#############################################################################
##
##  Input:
##    G     a finite p-group
##
##  Output:
##    The Galois order of G.
##
GaloisOrder := function( G )
    local p, s;

    # assert that G is a p-group
    if not IsPGroup( G ) then
        ErrorNoReturn( "Galois order is only defined for finite p-groups" );
    fi;

    # get prime and size of automorphism group
    p := PrimePGroup( G );
    s := Size( AutomorphismGroup( G ) );

    # return Galois order
    return Product( Filtered( Factors( s ), x -> x <> p ) );
end;

