#############################################################################
##                                                                         ##
##  compact.gi                      PTREE                  Alexander Cant  ##
##                                                                         ##
#############################################################################

#############################################################################
##
##  Input:
##    pt    a p-tree or the grp-entry of a p-tree
##
##  Output:
##    The compact version of pt.
##
CompactTree := function( pt )
    local t, ct, m, i, j, k, ind1, ind2, l1, l2, sub1, sub2, S;

    Print( " Computing compact form of p-tree...\n" );

    # check input and set up
    if IsRecord( pt ) and IsBound( pt.grp ) then
        if pt.cmp then
            return pt;
        fi;
        t := StructuralCopy( pt.grp );
    else
        t := StructuralCopy( pt );
    fi;
    if not IsList( t ) then
        ErrorNoReturn( "cannot handle input" );
    fi;
    m := Length( t );

    # simplify tree (from deepest levels to root)
    for i in Reversed( [1..m] ) do

        # compare groups pairwise by parent, Galois order and subtree
        for j in [1..Length( t[i] )] do

            # forget group code
            t[i][j][5] := 0;

            # check multiplicity
            if t[i][j][4] <> 0 then
                for k in [j+1..Length( t[i] )] do

                    # check multiplicity
                    if t[i][k][4] <> 0 then

                        # compare parent and Galois order
                        if t[i][j]{ [1,3] } = t[i][k]{ [1,3] } then

                            # construct subtrees
                            sub1 := GetSubtree( t, i, j );
                            sub2 := GetSubtree( t, i, k );
                            
                            # check if subtrees are isomorphic
                            if PTreesIsomorphic( sub1, sub2 ) then

                                # increase multiplicity in t[i][j]
                                t[i][j][4] := t[i][j][4] + t[i][k][4];

                                # set multiplicity of all descendants of
                                # t[i][k] to 0
                                t[i][k][4] := 0;
                                for l1 in [i+1..m] do
                                    for l2 in [1..Length( t[l1] )] do
                                        if IsDescendantOf( t, l1, l2, i, k )
                                                then
                                            t[l1][l2][4] := 0;
                                        fi;
                                    od;
                                od;
                            fi;
                        fi;
                    fi;
                od;
            fi;
        od;
    od;

    # remove groups with mult = 0 and fix indices of parents with GetSubtree
    S := GetSubtree( t, 1, 1 );

    # set up output
    if IsRecord( pt ) then
        ct := rec( );
        ct.p    := pt.p;
        ct.n    := pt.n;
        ct.root := pt.root;
        ct.dep  := pt.dep;
        ct.gal  := pt.gal;
        ct.info := pt.info;
        ct.cmp  := true;
        ct.grp  := S;
    else
        ct := S;
    fi;

    Print( " Done!\n\n" );

    return ct;
end;

