#############################################################################
##                                                                         ##
##  ptrees.gi                       PTREE                  Alexander Cant  ##
##                                                                         ##
#############################################################################

#############################################################################
##
##  Input:
##    pt    a p-tree containing at least one group
##    k     a positive integer
##
##  Output:
##    A p-tree that contains pt as well as all groups with distance at most
##    k to the root of pt and a Galois order that is divisible by some
##    element of pt.gal, i.e. an "extension" of pt.
##
ExtendDescendantTree := function( pt, k )
    local T, i, j, G, desc;

    # catch the cases where pt is compact or the last level is not complete
    if pt.cmp then
        ErrorNoReturn( "a compact p-tree cannot be extended" );
    fi;
    if not pt.info then
        ErrorNoReturn( "deepest level of given tree may be incomplete, ",
            "please check this and either set info to true or unbind ",
            "this level" );
    fi;

    # copy the given p-tree
    T := StructuralCopy( pt );

    # check if T.grp exists
    if not IsBound( T.grp ) then
        T.grp := [ ];
    fi;
    
    # try to compute the root if it is not given in T.grp
    if Length( T.grp ) = 0 then
        if IsBound( T.root ) then
            G := PcGroupCode( T.root, T.p^(T.dep+T.n) );
            T.grp[1][1] := [ 0, T.dep+T.n, GaloisOrder( G ), 1, T.root ];
        else
            ErrorNoReturn( "cannot handle empty p-tree" );
        fi;
    fi;

    # case that no new levels have to be computed
    if Length( T.grp ) > k + 1 then
        Print( " Nothing to do, all wanted levels already exist!\n" );
        return T;
    fi;

    # print info on used levels
    if Length( T.grp ) > 1 then
        Print( " Using levels 0 to ", Length( T.grp ) - 1,
            " from given p-tree...\n" );
    fi;

    # compute remaining levels
    for i in [Length(pt.grp)+1..k+1] do
        Print( " Computing level ", i-1, " of ", k, "...\n" );

        # create new level and set T.info to false
        T.grp[i] := [ ];
        T.info := false;

        # loop over the previous level
        for j in [1..Length( T.grp[i-1] )] do
            # get group represented by T.grp[i-1][j]
            G := PcGroupCode( T.grp[i-1][j][5], T.p^T.grp[i-1][j][2] );

            # compute capable immediate descendants of G
            desc := PqDescendants( G : StepSize := 1,
                                AllDescendants := false );

            # compute additonal information
            desc := List( desc, x -> [ j, LogInt( Size( x ), T.p ),
                                GaloisOrder( x ), 1, CodePcGroup( x ) ] );

            # keep only groups that are not on the main-line and whose Galois
            # order is divisible by some element of T.gal
            desc := Filtered( desc, x -> x[3] <> (T.p - 1)^2 and
                        ForAny( T.gal, y -> IsInt( x[3] / y ) ) );

            # sort new descendants by Galois order
            Sort( desc, function( x, y ) return x[3] > y[3]; end );

            # add new descendants to T.grp[i]
            Append( T.grp[i], StructuralCopy( desc ) );
        od;

        # set T.info to true
        T.info := true;

        # print info on number of new descendants
        Print( " Done! Got " );
        if Length( T.grp[i] ) = 1 then
            Print( "1 group.\n\n" );
        else
            Print( Length( T.grp[i] ), " groups.\n\n" );
        fi;

        # print info if the end was reached
        if i = k+1 and Length( T.grp[i] ) > 0 then
            Print( " Reached the desired depth.\n\n" );
        elif Length( T.grp[i] ) = 0 then
            Print( " ", i-2, " was the final depth. Exiting loop.\n\n" );
            break;
        fi;
    od;

    return T;
end;


#############################################################################
##
##  Input:
##    G     a non-trivial p-group of maximal class with p > 5
##    gal   a positive integer or a list of positive integers
##    dis   a non-negative integer
##
##  Output:
##    The subtree of the skeleton that contains G starting at G containing
##    all descendants of G that have a Galois order divisible by gal resp.
##    an element of gal and any distance to G, if dis = 0, or distance at
##    most dis to G, if dis > 0.
##
DescendantTree := function( G, gal, dis )
    local p, H, dep, lcs, n, pt, i, j;

    # set up gal
    if IsPosInt( gal ) then
        gal := [ gal ];
    elif IsList( gal ) and ForAll( gal, IsPosInt ) then
        gal := Set( gal );
    else
        ErrorNoReturn( "2nd arg. must be a (list of) positive integer(s)" );
    fi;

    # check dis
    if not IsInt( dis ) or dis < 0 then
        ErrorNoReturn( "3rd arg. must be a non-negative integer" );
    fi;

    # check G
    if not IsGroup( G ) or not IsPGroup( G ) or Size( G ) = 1 then
        ErrorNoReturn( "1st arg. must be a non-trivial finite p-group" );
    fi;

    # get prime and assert that p > 5 and that G has maximal class
    p := PrimePGroup( G );
    if p < 7 then
        ErrorNoReturn( "cannot handle p-groups with p < 7" );
    fi;
    if LogInt( Size( G ), p ) - 1 <> NilpotencyClassOfGroup( G ) then
        ErrorNoReturn( "1st arg. has to be a group of maximal class" );
    fi;

    # assert that G is a PcGroup
    if IsPcpGroup( G ) then
        G := PcpGroupToPcGroup( G );
    elif IsFpGroup( G ) then
        G := PcGroupFpGroup( G );
    elif not IsPcGroup( G ) then
        ErrorNoReturn( "group has to be a PcGroup, PcpGroup or FpGroup" );
    fi;

    # compute the index of the branch G is in and its depth in this branch
    # via repeatedly computing parents until the main-line is reached
    H := G;
    dep := 0;
    while GaloisOrder( H ) <> (p-1)^2 do
        lcs := LowerCentralSeries( H );
        H := H / lcs[Length(lcs)-1];
        dep := dep + 1;
    od;
    n := LogInt( Size( H ), p );

    # assert that G is in the skeleton
    if not dep in [0..n-2*p+8] or not IsCapable( G ) then
        ErrorNoReturn( "the given group is not contained in the skeleton" );
    fi;

    # if dis = 0, then set dis to the maximal distance to G in the skeleton
    if dis = 0 then
        dis := n-2*p+8 - dep;
    fi;

    # set up record pt except for pt.gal
    pt := rec( );
    pt.p := p;
    pt.n := n;
    pt.root := CodePcGroup( G );
    pt.dep := dep;
    pt.cmp := false;
    pt.grp := [ [ [ 0, LogInt( Size( G ), p ), GaloisOrder( G ), 1,
                    CodePcGroup( G ) ] ] ];
    pt.info := true;

    # check if gal contains any divisor of Gal(G)
    gal := Filtered( gal, x -> IsInt( GaloisOrder( G ) / x ) );
    if IsEmpty( gal ) then
        pt.gal := [ ];
        return pt;
    fi;

    # minimize gal and add it to pt
    for i in [1..Length( gal )] do
        if IsBound( gal[i] ) then
            for j in [i+1..Length( gal )] do
                if IsBound( gal[j] ) and IsInt( gal[j] / gal[i] ) then
                    Unbind( gal[j] );
                fi;
            od;
        fi;
    od;
    gal := Flat( gal );
    pt.gal := gal;

    # use ExtendDescendantTree for the computation of wanted descendants
    pt := ExtendDescendantTree( pt, Minimum( n-2*p+8 - dep, dis ) );

    return pt;
end;


#############################################################################
##
##  Input:
##    p    a prime > 5
##    n    an integer > 2*p-8
##    gal  a positive integer or a list of positive integers
##    dis  a non-negative integer
##
##  Output:
##    The subtree of the n-th skeleton of p-groups containing all groups that
##    have a Galois order divisible by gal resp. an element of gal and any
##    distance to the main-line, if dis = 0, or distance at most dis to the
##    main-line, if dis > 0.
##
Skeleton := function( p, n, gal, dis )
    local G, i, j;

    # check p and n
    if not ( IsInt( p ) and IsPrime( p ) and p > 5 ) then
        ErrorNoReturn( "1st arg. has to be a prime > 5" );
    fi;
    if not ( IsInt( n ) and n > 2*p-8 ) then
        ErrorNoReturn( "2nd arg. has to be an integer > 2*p-8 = ", 2*p-8 );
    fi;

    # check and adjust dis
    if not IsInt( dis ) or dis < 0 then
        ErrorNoReturn( "4th arg. must be a non-negative integer" );
    fi;
    if dis = 0 or dis > n-2*p+8 then
        dis := n-2*p+8;
    fi;

    # set up gal
    if IsPosInt( gal ) then
        gal := [ gal ];
    elif IsList( gal ) and ForAll( gal, IsPosInt ) then
        gal := Set( gal );
    else
        ErrorNoReturn( "3rd arg. must be a (list of) positive integer(s)" );
    fi;

    # clean and minimize gal
    gal := Filtered( gal, x -> IsInt( (p-1) / x ) );
    for i in [1..Length( gal )] do
        if IsBound( gal[i] ) then
            for j in [i+1..Length( gal )] do
                if IsBound( gal[j] ) and IsInt( gal[j] / gal[i] ) then
                    Unbind( gal[j] );
                fi;
            od;
        fi;
    od;
    gal := Flat( gal );

    # print information on situation
    Print( "\n Prime:               ", p, "\n" );
    Print( " Branch:              ", n, "\n" );
    Print( " Max. level computed: ", dis, "\n" );
    Print( " Depth of skeleton:   ", n-2*p+8, "\n" );
    if 1 in gal then
        Print( " Galois orders:       all\n\n" );
    elif p-1 in gal then
        Print( " Galois order:        ", p-1, "\n\n" );
    else
        Print( " Galois orders:       multiples of " );
        for i in [1..Length( gal )] do
            Print( gal[i] );
            if i = Length( gal ) - 1 then
                Print( " or " );
            elif i < Length( gal ) then
                Print( ", " );
            fi;
        od;
        Print( "\n\n" );
    fi;


    # get main-line group
    Print( " Computing main-line group...\n" );
    G := MainLineGroup( p, n );
    Print( " Done!\n\n" );

    # compute the skeleton using DescendantTree
    return DescendantTree( G, gal, dis );
end;

