#############################################################################
##                                                                         ##
##  tree2latex.gi                   PTREE                  Alexander Cant  ##
##                                                                         ##
#############################################################################

#############################################################################
##
##  Input:
##    pt    a p-tree or the grp-entry of a p-tree
##    fn    a string
##
##  Output:
##    None, but a tikz-picture displaying pt is printed to the file "fn.tex"
##    in DIRTEX.
##
TreeToLaTeX := function( pt, fn )
    local file, i, j;

    # check input
    if IsList( pt ) then
        TreeToLaTeX( rec( grp := pt ) );
        return;
    fi;
    if not IsRecord( pt ) or not IsString( fn ) or IsEmpty( fn ) then
        ErrorNoReturn( "cannot handle the input" );
    fi;
    
    # step 0: set up
    file := Filename( DIRTEX, Concatenation( fn, ".tex" ) );

    # check if file exists
    if IsExistingFile( file ) then
        Error( "file already exists, 'return;' if it may be replaced" );
    fi;

    Print( " Printing p-tree to ", file, "...\n" );
    
    # step 1: print preamble and tikz-header
    PrintTo( file, "" );
    if NumberOfLeavesOfDescendantTree( pt, 1, 1 ) <= 27 then
        AppendTo( file, "\\documentclass[a4paper,10pt]{article}\n" );
    else
        AppendTo( file,
            "\\documentclass[a4paper,landscape,10pt]{article}\n" );
    fi;    
    AppendTo( file, "\\usepackage[utf8x]{inputenc}\n" );
    AppendTo( file, "\\usepackage{tikz}\n" );
    AppendTo( file, "\\usepackage{verbatim}\n" );
    AppendTo( file, "\\usepackage{ifthen}\n" );
    AppendTo( file,
        "\\usepackage[right=7mm,left=7mm, top=10mm, bottom=10mm]{geometry}\n" );
    AppendTo( file, "\\usetikzlibrary{arrows,shapes,positioning}\n\n" );
    AppendTo( file, "\\begin{document}\n\n" );
    AppendTo( file,
        "\\tikzstyle{vert}=[draw,circle,line width=1pt,minimum size=10pt,inner sep=0pt]\n" );
    AppendTo( file, "\\tikzstyle{edge} = [draw,>=latex,->]\n" );
    AppendTo( file, "\\tikzstyle{dashededge} = [draw,dashed,>=latex,-]\n" );
    AppendTo( file, "\\tikzstyle{weight} = [font=\\small]\n\n" );
    AppendTo( file, "\\thispagestyle{empty}\n\n" );

    # step 2: print information if pt.p, pt.n and pt.dep are bound
    if IsBound( pt.p ) and IsBound( pt.n ) and IsBound( pt.dep ) then
        AppendTo( file, "\\noindent $p = " );
        AppendTo( file, String( pt.p ) );
        AppendTo( file, "$, branch = $" );
        AppendTo( file, String( pt.n ) );
        if 1 in pt.gal and Length( pt.grp ) = pt.n - 2*pt.p+8 + 1 and pt.info then
            # case: p-tree is a full skeleton
            AppendTo( file, "$, full skeleton \\vspace*{5mm}\n\n" );
        elif pt.dep = 0 then
            # case: not a full skeleton but root is on the main-line
            AppendTo( file, "$, root on the main-line \\vspace*{5mm}\n\n" );
        else
            # case: root not on the main-line
            AppendTo( file, "$, root of size $" );
            AppendTo( file, String( pt.p ) );
            AppendTo( file, "^{" );
            AppendTo( file, String( pt.n + pt.dep ) );
            AppendTo( file, "}$ \\vspace*{5mm}\n\n" );
        fi;
    fi;

    # step 3: code for the tree
    AppendTo( file, "\\begin{figure}[h]\n" );
    AppendTo( file, "\\begin{tikzpicture}[xscale=0.7]\n\n" );
    
    # step 3.1: generate a list of vertices
    AppendTo( file,
        "    % First we draw the vertices (including Gal.ord. and mult.)\n" );
    AppendTo( file, "    \\foreach \\pos / \\name / \\gal / \\mult in\n" );
    AppendTo( file, "    {" );
    # a vertex at position (x,y) with name name that represents (a) group(s)
    # with Galois order gal and multiplicity mult is generated by
    # {(x,y)/name/gal/mult}
    for i in [1..Length( pt.grp )] do
        for j in [1..Length( pt.grp[i] )] do
            AppendTo( file, Concatenation( "{(",
                String( GetAbs( pt, i, j ) ), ",", String( -i ), ")/v-i",
                String( i ), "-j", String( j ), "/",
                String( pt.grp[i][j][3] ), "/",
                String( pt.grp[i][j][4] ), "}" ) );
            if not ( i = Length( pt.grp ) and j = Length( pt.grp[i] ) ) then
                AppendTo( file, ", " );
                if j = Length( pt.grp[i] ) then
                    AppendTo( file, "\n     " );
                fi;
            fi;
        od;
    od;
    AppendTo(file, "}\n");
    
    # step 3.2: print code to print vertices
    AppendTo( file, "    {\n    \\ifthenelse{\\mult > 1}\n" );
    AppendTo( file,
        "        { \\node[vert, label={[shift={(0.33,-0.4)}]\\footnotesize$\\mult$}] (\\name) at \\pos {\\small\\gal}; }\n");
    AppendTo( file, "        { \\node[vert] (\\name) at \\pos {\\small\\gal}; }\n}\n" );
    
    # step 3.3: if pt.n is bound and pt.dep = 0, print a name for the root
    if IsBound( pt.n ) and IsInt( pt.n ) and IsBound( pt.dep ) and pt.dep = 0 then
        AppendTo( file,
            Concatenation( "      \\node[left=0.1cm of v-i1-j1] {$S_{",
            String( pt.n ), "}$};\n\n" ) );
    fi;

    # step 3.4: generate a list of edges
    AppendTo( file, "    % Connect vertices with edges\n" );
    AppendTo( file, "    \\foreach \\source / \\dest in\n" );
    AppendTo( file, "    {" );
    # an edge from name1 to name2 is generated by "name1/name2"
    for i in [2..Length( pt.grp )] do
        for j in [1..Length( pt.grp[i] )] do
            AppendTo( file, Concatenation( "v-i", String(i-1), "-j",
                String( pt.grp[i][j][1] ), "/v-i", String( i ), "-j",
                String( j ) ) );
            if not ( i = Length( pt.grp ) and j = Length( pt.grp[i] ) ) then
                AppendTo( file, ", " );
                if j = Length( pt.grp[i] ) then
                    AppendTo( file, "\n     " );
                fi;
            fi;
        od;
    od;
    AppendTo( file, "}\n" );
    
    # step 3.5: print code to print edges
    AppendTo( file, "    \\path[edge] (\\source) -- (\\dest);\n" );
    
    # step 4: print foot
    AppendTo( file, "\\end{tikzpicture}\n" );
    AppendTo( file, "\\end{figure}\n\n" );
    AppendTo( file, "\\end{document}\n" );
    
    Print( " Done!\n\n" );
end;

