#############################################################################
##                                                                         ##
##  treetools.gi                    PTREE                  Alexander Cant  ##
##                                                                         ##
#############################################################################

#############################################################################
##
##  Input:
##    t1, t2    each a p-tree or the grp-entry of a p-tree
##
##  Output:
##    A Boolean indicating whether or not t1 and t2 are isomorphic as vertex
##    colored trees, where each vertex is colored with a Galois order and
##    each but the root with a multiplicity.
##
##  This method is based on the algorithm on page 84 of [13].
##
PTreesIsomorphic := function( t1, t2 )
    local T1, T2, L1, L2, LL, S1, S2, i, j, k, m, ind1, ind2, v;

    # set up
    if IsRecord( t1 ) then
        T1 := StructuralCopy( t1.grp );
    else
        T1 := StructuralCopy( t1 );
    fi;
    if IsRecord( t2 ) then
        T2 := StructuralCopy( t2.grp );
    else
        T2 := StructuralCopy( t2 );
    fi;

    # check input
    if not IsList( T1 ) or not IsList( T2 ) then
        ErrorNoReturn( "cannot handle input" );
    fi;

    # determine depth of the tree
    m := Length( T1 );

    # set multiplicity on level m-1 to 0 (irrelevant)
    T1[1][1][4] := 0;
    T2[1][1][4] := 0;

    #  2: assign level numbers to all nodes of T1 and T2
    #     (root: index 1, level m-1; at maximal depth: index m, level 0)
    #      index = m - level; level = m - index
    if Length( T1 ) <> Length( T2 ) then
        return false;
    elif Length( T1 ) = 1 then
        return T1[1][1]{ [3,4] } = T2[1][1]{ [3,4] };
    elif Length( T1 ) = 0 then
        return true;
    fi;
    for i in [1..Length( T1 )] do
        if Length( T1[i] ) <> Length( T2[i] ) then
            return false;
        fi;
        for j in [1..Length( T1[i] )] do
            # assign level numbers
            T1[i][j][6] := m-i;
            T2[i][j][6] := m-i;
            # assign tuples
            T1[i][j][8] := [ T1[i][j][3], T1[i][j][4] ];  # Difference to
            T2[i][j][8] := [ T2[i][j][3], T2[i][j][4] ];  # TreeIsomorphic!
        od;
    od;
    
    #  3: assign to all leaves of T1 and T2 the integer 0
    for i in Reversed( [3..m] ) do
        ind1 := [];
        ind2 := [];
        for j in [1..Length( T1[i] )] do
            Append( ind1, [ T1[i][j][1] ] );
            Append( ind2, [ T2[i][j][1] ] );
        od;
        for j in [1..Length( T1[i-1] )] do
            if not j in ind1 then
                T1[i-1][j][7] := 0;
            fi;
            if not j in ind2 then
                T2[i-1][j][7] := 0;
            fi;
        od;
    od;
    for j in [1..Length( T1[m] )] do
        T1[m][j][7] := 0;
        T2[m][j][7] := 0;
    od;
    
    #  4: let L1 be a list of the leaves of T1 at level 0
    #  5: let L2 be a list of the leaves of T2 at level 0
    L1 := [];
    L2 := [];
    for i in [1..Length( T1[m] )] do
        L1[i] := StructuralCopy( T1[m][i] );
        L2[i] := StructuralCopy( T2[m][i] );
    od;
    
    #  6: for all levels i starting from 1 do
    #  7:     << assign integers to all nodes at level i >>
    for i in Reversed( [1..m-1] ) do
        # !!! not in the algorithm !!!
        Sort(L1);
        Sort(L2);
        
        #     << assign integers to all nodes at level i >>:
        #  1: for all nodes v in list L1 do
        #  2:   assign to the next component of the tuple associated with the
        #           parent of v the integer assigned to v
        for v in L1 do
            Append( T1[m-v[6]-1][v[1]][8], [v[7],v[3],v[4]] );
        od;
        for v in L2 do
            Append( T2[m-v[6]-1][v[1]][8], [v[7],v[3],v[4]] );
        od;
        
        #  3: let S1 be the sequence of tuples created for the nonleaves of
        #           T1 on level i
        #  4: let S2 be the corresponding sequence of tuples of T2
        S1 := [];
        S2 := [];
        for j in [1..Length( T1[i] )] do
            if not IsBound( T1[i][j][7] ) then
                Append( S1, [T1[i][j][8]] );
            elif T1[i][j][7] <> 0 then
                Append( S1, [T1[i][j][8]] );
            fi;
            if not IsBound( T2[i][j][7] ) then
                Append( S2, [T2[i][j][8]] );
            elif T2[i][j][7] <> 0 then
                Append( S2, [T2[i][j][8]] );
            fi;
        od;
        
        #  5: bucket sort S1
        #  6: bucket sort S2
        Sort( S1 );
        Sort( S2 );
        
        #  7: if S1 <> S2 then
        #  8:   T1 and T2 are not isomorphic
        if S1 <> S2 then
            return false;
        
        #  9: else
        else
            # 10:   let L1 be an empty list of nodes
            # 11:   for all k from 1 to the number of distinct tuples on S1
            #               do
            # 12:     for all nodes v of T1 on level i represented by the
            #               k-th distinct tuple on S1 do
            # 13:       assign to node v the integer k
            # 14:       append node v to L1
            # 15:   append to the front of L1 all leaves of T1 on level i
            L1 := [];
            S1 := Set( S1 );
            for k in [1..Length( S1 )] do
                for v in T1[i] do
                    if v[8] = S1[k] then
                        v[7] := k;
                        Append( L1, [v] );
                    fi;
                od;
            od;
            for v in T1[i] do
                if IsBound( v[7] ) then
                    if v[7] = 0 then
                        LL := [v];
                        Append( LL, L1 );
                        L1 := StructuralCopy( LL );
                    fi;
                fi;
            od;

            # 16:   let L2 be the corresponding list of nodes of T2
            L2 := [];
            S2 := Set( S2 );
            for k in [1..Length( S2 )] do
                for v in T2[i] do
                    if v[8] = S2[k] then
                        v[7] := k;
                        Append( L2, [v] );
                    fi;
                od;
            od;
            for v in T2[i] do
                if IsBound( v[7] ) then
                    if v[7] = 0 then
                        LL := [v];
                        Append( LL, L2 );
                        L2 := StructuralCopy( LL );
                    fi;
                fi;
            od;
        fi;
    od;
    
    #  8: if the roots of T1 and T2 are assigned the same integer then
    #  9:     T1 and T2 are isomorphic
    # 10: else
    # 11:     T1 and T2 are not isomorphic
    return T1[1][1][7] = T2[1][1][7];
end;


#############################################################################
##
##  Input:
##    pt    a p-tree or the grp-entry of a p-tree
##    a, b  positive integers
##
##  Output:
##    The full subtree of T starting at T[a][b].
##
GetSubtree := function( pt, a, b )
    local T, S, ind1, ind2, i, j, k, l, st;

    # check input
    if IsRecord( pt ) and IsBound( pt.grp ) then
        T := pt.grp;
    else
        T := pt;
    fi;
    if not IsList( T ) or not IsPosInt( a ) or not IsPosInt( b )
            or not IsBound( T[a] ) or not IsBound( T[a][b] ) then
        ErrorNoReturn( "cannot handle input" );
    fi;

    # set up subtree S with root T[a][b]
    S := [];
    S[1] := [ StructuralCopy( T[a][b] ) ];
    S[1][1][1] := 0;

    # set up lists ind1 and ind2 that are used to fix the parent indices in S
    # (if T[i][j] is in S and its parent in T is T[i-1][ind1[1][k]], then the
    #   parent of T[i][j] in S is S[i-a+1][ind1[2][k]])
    ind1 := [ [b], [1] ];
    ind2 := [ [], [] ];
    
    # add the elements of T to S that are descendants of T[a][b]
    for i in [a+1..Length( T )] do

        # generate an empty list
        S[i-a+1] := [];

        # for each vertex in T at level i do
        for j in [1..Length( T[i] )] do
            for k in [1..Length( ind1[1] )] do

                # if the parent of T[i][j] is a descendant of T[a][b] and
                # the multiplicity of T[i][j] is greater than 0 then
                if T[i][j][1] = ind1[1][k] and T[i][j][4] > 0 then

                    # add T[i][j] to S
                    Add( S[i-a+1], StructuralCopy( T[i][j] ) );

                    # fix the index of the parent of T[i][j] in S
                    l := Length( S[i-a+1] );
                    S[i-a+1][l][1] := ind1[2][k];

                    # add old and new index of T[i][j] to ind2
                    Add( ind2[1], j);
                    Add( ind2[2], l );
                fi;
            od;
        od;

        # update ind1 to ind2 and reset ind2
        ind1 := StructuralCopy( ind2 );
        ind2 := [ [], [] ];
    od;

    # set up output
    if IsRecord( pt ) then
        st := rec( );
        st.p    := pt.p;
        st.n    := pt.n;
        st.root := pt.grp[a][b][5];
        st.dep  := pt.dep + a - 1;
        st.gal  := pt.gal;
        st.info := pt.info;
        st.cmp  := pt.cmp;
        st.grp  := S;
    else
        st := S;
    fi;

    return st;
end;


#############################################################################
##
##  Input:
##    pt            a p-tree or the grp-entry of a p-tree
##    a, b, c, d    positive integers
##
##  Output:
##    A Boolean indicating whether or not pt.grp[a][b] (resp. pt[a][b]) is a
##    descendant of pt.grp[c][d] (resp. pt[c][d]) in pt.grp (resp. pt).
##
IsDescendantOf := function( pt, a, b, c, d )
    local i, curr;

    # check input
    if IsList( pt ) then
        return IsDescendantOf( rec( grp := pt ), a, b, c, d );
    elif not IsRecord( pt ) then
        ErrorNoReturn( "1st arg. has to be a record or a list" );
    fi;
    if not ForAll( [ a, b, c, d ], IsPosInt ) then
        ErrorNoReturn( "2nd to 5th arg. have to be positive integers" );
    fi;
    if not IsBound( pt.grp[a] ) or not IsBound( pt.grp[a][b] ) then
        ErrorNoReturn( "no vertex at first position" );
    fi;
    if not IsBound( pt.grp[c] ) or not IsBound( pt.grp[c][d] ) then
        ErrorNoReturn( "no vertex at second position" );
    fi;

    # compute iteratively parents of pt.grp[a][b] until at level c
    if c < a then
        curr := pt.grp[a][b];
        for i in Reversed( [c+1..a-1] ) do
            curr := pt.grp[i][curr[1]];
        od;
        return curr[1] = d;
    elif a = c then
        return b = d;
    else
        return false;
    fi;
end;


#############################################################################
##
##  Input:
##    pt    a p-tree or the grp-entry of a p-tree
##    a, b  positive integers
##
##  Output:
##    The number of leaves of the full descendant tree of pt.grp[a][b]
##    (resp. pt[a][b]) in pt.grp (resp. pt).
##
NumberOfLeavesOfDescendantTree := function( pt, a, b )
    local w, i;

    # check input
    if IsList( pt ) then
        return NumberOfLeavesOfDescendantTree( rec( grp := pt ), a, b );
    elif not IsRecord( pt ) then
        ErrorNoReturn( "1st arg. has to be a record or a list" );
    fi;
    if not IsPosInt( a ) or not IsPosInt( b ) then
        ErrorNoReturn( "2nd and 3rd arg. have to be positive integers" );
    fi;
    if not IsBound( pt.grp[a] ) or not IsBound( pt.grp[a][b] ) then
        ErrorNoReturn( "no vertex at given position" );
    fi;

    # compute the number of leaves inductively
    w := 0;
    if IsBound( pt.grp[a+1] ) then
        for i in [1..Length( pt.grp[a+1] )] do
            if pt.grp[a+1][i][1] = b then
                w := w + NumberOfLeavesOfDescendantTree( pt, a+1, i );
            fi;
        od;
    fi;
    if w = 0 then
        w := 1;
    fi;

    return w;
end;


#############################################################################
##
##  Input:
##    pt    a p-tree or the grp-entry of a p-tree
##    a, b  positive integers
##
##  Output:
##    The abscissa (i.e. the x-value) of the vertex representing pt.grp[a][b]
##    (resp. pt[a][b]) in the printed version of pt.
##
GetAbs := function( pt, a, b )
    local x, i;

    # check input
    if IsList( pt ) then
        return GetAbs( rec( grp := pt ), a, b );
    elif not IsRecord( pt ) then
        ErrorNoReturn( "1st arg. has to be a record or a list" );
    fi;
    if not IsPosInt( a ) or not IsPosInt( b ) then
        ErrorNoReturn( "2nd and 3rd arg. have to be positive integers" );
    fi;
    if not IsBound( pt.grp[a] ) or not IsBound( pt.grp[a][b] ) then
        ErrorNoReturn( "no vertex at given position" );
    fi;

    # compute abscissa inductively
    if a = 1 then
        x := 1;
    else
        x := GetAbs( pt, a-1, pt.grp[a][b][1] );
        for i in [1..b-1] do
            if pt.grp[a][i][1] = pt.grp[a][b][1] then
                x := x + NumberOfLeavesOfDescendantTree( pt, a, i );
            fi;
        od;
    fi;
    
    return x;
end;

